#!/usr/bin/perl

use strict;
use warnings;
use CGI;
use CGI::Session;
use MIME::Base64;
use Crypt::Blowfish;
use Crypt::CBC;
use bytes;
use URI::Escape;
use LWP::UserAgent;
use HTTP::Request;

# {{{ BEGIN CONFIGURATION SECTION

# where credentials are posted
use constant POST_URL => "http://localhost/gw/webacc";
# where to send the client if the logon succeeeds
use constant TARGET_URL => "https://mymail.messiah.edu/gw/webacc";
# where to send the client if the logon failed
use constant ERROR_URL => "https://mymail.messiah.edu/gw/webacc";

# secret key... must be 16 numbers, preferably random, each between
# -128 and 127. IMPORTANT- the numbers in sso_login and webacc-helper.cgi
# must MATCH. And if you have multiple servers, each server must have
# the same set of numbers.

my @KEY_BYTES = qw(
	60 -81 32 -31 3 103 30 -73
	25 59 51 -78 26 -13 25 35
	);

# }}} END CONFIGURATION SECTION

my $CGI = new CGI;
my $CGI_Session = new CGI::Session;

if (my $ciphertext = $CGI->param("k") || $CGI->cookie("k"))
{
	my $credentials = decrypt($ciphertext);
	my ($user, $pass) = split(/:/, $credentials, 2);
	unless ($pass)
	{
		print $CGI->header(-status => 500);
		print qq|<html><body>
		An error occurred accessing GroupWise WebAccess.
		Please try again later.
		If the problem persists, contact ITS.
		<!-- k=$ciphertext -->
		</body></html>\n|;
		exit;
	}
	my $auth = {
		username => $user,
		password => $pass,
		};
	if (my $context = perform_webacc_login($auth))
	{
		redirect_into_application($context);
		exit 0;
	}
	else
	{
		print $CGI->redirect(ERROR_URL);
	}
}
else
{
	# this is a bit of (optional) trickery....
	# if no k= cookie is present, redirect to the CPIP URL
	# that will perform the login
	$ENV{QUERY_STRING} ||= "merge=webacc";
	my $dest_url = TARGET_URL . "?" . $ENV{QUERY_STRING};
	my $url = "http://mcsquare.messiah.edu/cp/ip/login?sys=gw7&url="
		. uri_escape($dest_url);
	print $CGI->redirect($url);
}
exit;

sub redirect_into_application
{
	my ($context) = @_;

	my $query_string = $ENV{QUERY_STRING};
	my @parts = split(/&/, $query_string);
	@parts = grep !/^User\.context=/, @parts;
	push @parts, "User.context=" . uri_escape($context->{id});
	$query_string = join("&", @parts);

	my $dest_url = TARGET_URL . "?" . $query_string;

	my $cookies = $context->{cookies};
	my $cookie_header = join("; ",
		map {
			my $name = $_;
			my ($value, undef) = split /;/, $cookies->{$name}, 2;
			"$name=$value"
		} keys %$cookies);

	print $CGI->redirect(
		-uri => $dest_url,
		-cookie => $cookie_header,
		);
}

sub perform_webacc_login
{
	my ($auth) = @_;

	my %cookies;
	my $ua = LWP::UserAgent->new;
	if (my $user_agent_string = $ENV{HTTP_USER_AGENT})
	{
		# this is important to make sure the browser-specific
		# hacks in WebAccess are enabled
		$ua->agent($user_agent_string);
	}

	# now post credentials
	my $req2 = HTTP::Request->new(POST => POST_URL);
	$req2->content_type("application/x-www-form-urlencoded");
	$req2->content(join('&',
			"User.id=" . uri_escape($auth->{username}),
			"User.password=" . uri_escape($auth->{password}),
			"action=User.Login",
			"error=login",
			"merge=webacc",
			"Low.bandwidth=0",
			"User.displayDraftItems=1",
			"User.interface=css",
			"User.Theme.index=1",
			"User.settings.speed=high",
			"User.settings.save=on",
			"Url.hasJavaScript=1",
			));

#	print STDERR "Ok, I'm posting the login request:\n";
#	print STDERR $req2->as_string;

	my $res2 = $ua->request($req2);

#	print STDERR "Got this response:\n";
#	print STDERR $res2->as_string;

	unless ($res2->is_success)
	{
		warn "got bad response: " . $res2->status_line . "\n";
		return;
	}

	my $content = $res2->content;
	unless ($content =~ /FRAMESET/)
	{
		warn "FRAMESET not found in response\n";
		return;
	}

	my $sessionid;
	if ($content =~ /User\.context=([^"&]+)/)
	{
		my $context_id = $1;

		# check whether GroupWise gave me any cookies
		my $cookies = {};
		foreach my $c ($res2->header("Set-Cookie"))
		{
			my ($name, $value) = split /=/, $c, 2;
			$cookies->{$name} = $value;
		}

		# successful login
		return {
			id => $context_id,
			cookies => $cookies,
			};
	}
	else
	{
		warn "User.context not found in response\n";
		return;
	}
}

# takes base64-encoded encrypted string, returns username:password
sub decrypt
{
	my $ciphertext = decode_base64(shift);
	my $key = join("", map { chr($_) } @KEY_BYTES);
	my $cipher = Crypt::CBC->new(
			-key => $key,
			-cipher => "Blowfish",
			);
	return $cipher->decrypt($ciphertext);
}

# takes username:password, returns base64-encoded encrypted string
sub encrypt
{
	my $plaintext = shift;

	my $key = join("", map { chr($_) } @KEY_BYTES);
	my $cipher = Crypt::CBC->new(
			-key => $key,
			-cipher => "Blowfish",
			);
	my $ciphertext = $cipher->encrypt($plaintext);
	return encode_base64($ciphertext, "");
}
